# Module variables from the Config.psd1 file
$ModuleParent = Split-Path -path ($(Get-Item -Path $MyInvocation.MyCommand.Source).FullName) -Parent
if ($PSVersionTable.PSVersion -eq [version]"2.0") {
    Import-LocalizedData -BindingVariable "Config" -FileName "Config.psd1"
}
else {
    $Config = Import-PowerShellDataFile -Path (Join-Path -Path $ModuleParent -ChildPath "Config.psd1")
}

# Local to the module, not exported.
# Global because we want to increment this value from every module memeber as the log gets written to
$global:LogTailBegin = 0


function Confirm-PathExists {
    [CmdletBinding()]
    param (
        [Parameter()]
        [String]
        $Path
    )

    if (!(Test-Path -Path $Path)) {
        $PathParent = Split-Path -Path $Path -Parent
        if (!(Test-Path -Path $PathParent)) {
            New-Item -Path $PathParent -ItemType Directory -Force
        }
        New-Item -Path $Path -ItemType File -Force
    }
}

function Write-Log {
    # This particular log format is that of the NCR RPOS installer logs.
    # Format as per the InstallRPOS.exe.config spec
    # LogSpec = "%d{yyyy-MM-ddTHH:mm:sss} ;; %p ;; %t ;; %c ;; %m%n"
    # Or... 2023-02-11T15:19:01 ;; ERROR ;; [main] ;; RPOSInstaller ;; This is the log message \n <-(newline character)   

    [CmdletBinding()]
    param (
        [Parameter()]
        [String]
        $Path=$Config.RPOSLocalLogPath,
        [Parameter()]
        [ValidateSet("debug","info","warning","error","critial")]
        [String]
        $Level="error",
        [Parameter(Mandatory)]
        [String]
        $Message,
        [Parameter()]
        [String]
        $Thread="AUF-RPOS Function",
        [Parameter()]
        [String]
        $Logger="AUF-RPOS:Write-Log"

    )

    # Create the log file if necessary
    Confirm-PathExists -Path

    # Compose log message
    $Timestamp = $(Get-Date -format "yyyy-MM-dd;HH:mm:sss")
    $LogLine = ("{0} ;; {1} ;; [{2}] ;; {3} ;; {4}" -f @($Timestamp, $Level, $Thread, $Logger, $Message))

    Add-Content -Path $Path -Value $LogLine -Encoding Ascii
}

function Format-LogToCSV {
    [CmdletBinding()]
    param (
        [Parameter()]
        [string]
        $LogLines
    )
    # Parse logs for attributes (differs from log to log, so this function can be duplicated and changed for each purpose)
    # This particular log format is that of the NCR RPOS installer logs.
    # Format as per the InstallRPOS.exe.config spec
    # LogSpec = "%d{yyyy-MM-ddTHH:mm:sss} ;; %p ;; %t ;; %c ;; %m%n"
    # Or... 2023-02-11T15:19:01 ;; ERROR ;; [main] ;; RPOSInstaller ;; This is the log message \n <-(newline character)       
    foreach ($line in $LogLines) {
        $ContentSplit = $line.Split(" ;; ")

        $LogLineObject = [PSCustomObject]@{
            Level = $ContentSplit[1]
            Thread = $ContentSplit[2].trim("[]")
            Computer = $ENV:COMPUTERNAME
            TimeStamp = $ContentSplit[0]
            Logger = $ContentSplit[3]
            Message = $ContentSplit[4]
        }

        $LogList += $LogLineObject
    }

    return $LogList
}
function Export-Log {
    [CmdletBinding()]
    param (
        [Parameter()]
        [String]
        $Path=$Config.RPOSLocalLogPath,
        [Parameter()]
        [String]
        $RemotePath=$Config.RPOSRemoteLogPath,
        [Parameter()]
        [switch]
        $NoTail=$false
    )
    # Create the csv file if necessary
    Confirm-PathExists -Path $RemotePath

    if ($NoTail) {
        # Just read in all the logs, not a slice
        $LogLines = Get-Content -Path $Path
        $LogList = Format-LogToCSV -LogLines $LogLines
        $LogList | Export-Csv -Path $RPOSRemoteLogPath -NoClobber -Encoding ASCII -Append
    }
    else {
        if ($global:LogTailBegin) {
            # If the counter is not zero, read in the last $LineCounter lines
            $LogLines = Get-Content -Path $Path -TotalCount -($LineCounter)
        }
        else {
            # If the counter is 0, read in the whole thing
            $LogLines = Get-Content -Path $Path
        }
        
        # $SlicedLog = $LogLines[$global:LogTailBegin..-1]
        # $SliceLength = $SlicedLog.Length
        $SliceLength = $LogLines.Length
    
        $LogList = Format-LogToCSV -LogLines $LogLines
        $LogList | Export-Csv -Path $RPOSRemoteLogPath -NoClobber -Encoding ASCII -Append
    
        $global:LogTailBegin += $SliceLength
    }
}

function Start-ExeInstaller {
    [CmdletBinding()]
    param (
        [Parameter()]
        [String]
        $RPOSSetupPath=$Config.DefaultSetupDir
    )
    
    $Config.ModuleParent = Split-Path -path ($(Get-Item -Path $MyInvocation.MyCommand.Source).FullName) -Parent
    $InstallerPath = (Join-Path $RPOSSetupPath -ChildPath "Bin\InstallRPOS.exe")
    $SourceInstallerConfig = (Join-Path -Path $Config.ModuleParent -ChildPath "InstallRPOS.exe.config")
    $DestinationInstallerConfig = (Join-Path $RPOSSetupPath -ChildPath "Bin\InstallRPOS.exe.config")

    if (!(Test-Path -Path $SourceInstallerConfig)) {
        throw [System.IO.FileNotFoundException]
        exit 2
    }

    if (!(Test-Path -Path $DestinationInstallerConfig)) {
        throw [System.IO.FileNotFoundException]
        exit 2
    }

    Copy-Item -Path $SourceInstallerConfig -Destination $DestinationInstallerConfig -Force

    try {
        if (!(Test-Path -Path $InstallerPath)) {
            throw [System.IO.FileNotFoundException]
        }

        $InstallerProcess = (Start-Process -FilePath $InstallerPath)
    }
    catch [System.IO.FileNotFoundException] {
        $Message = ("$Installer path not found. Verify that file exists.")
        Write-Log -Message $Message -Thread $MyInvocation.MyCommand.Name
    }
    finally {
        while (!($InstallerProcess.HasExited)) {
            Start-Sleep -Seconds 5
            Export-Log
        }
        Export-Log
    }
}

function Get-InstalledSoftware {
    [CmdletBinding()]
    param (
        [Parameter()]
        [string[]]
        $Programs=""
    )
    # array to store objects in
    $ProgramList = @()

    $RegKeys = @(
        "HKLM:\Software\Microsoft\Windows\CurrentVersion\Uninstall\*",
        "HKLM:\SOFTWARE\WOW6432Node\Microsoft\Windows\CurrentVersion\Uninstall\*"
    )

    # Contents of registry keys for both bitnesses
    foreach ($Key in $RegKeys) {
        $Bitness = "x86" 
        if ($Key -eq "HKLM:\SOFTWARE\WOW6432Node\Microsoft\Windows\CurrentVersion\Uninstall\*") {
            $Bitness = "x64"
        }
        $RegValues = Get-ItemProperty $Key -Exclude "{*" | select-object DisplayName, InstallDate, DisplayVersion, EstimatedSize, Publisher, UninstallString

        foreach ($Item in $RegValues) {
            $Program = [PSCustomObject]@{
                Name = $Item.DisplayName
                Publisher = $Item.Publisher
                DateInstalled = $Item.InstallDate
                Version = $Item.DisplayVersion
                EstimatedSize = $Item.EstimatedSize
                UninstallPath = $Item.UninstallString
                Bitness = $Bitness
            }
            
            if ($ProgramList.Name -notcontains $Program.Name -and ($null -ne $Program.Name)) {
                $ProgramList += $Program
            }
        }
    }

    
    # Filter the list of installed programs to include only those specified by the user
    $OutputList = $ProgramsList | Where-Object -FilterScript {$Programs -contains $_.Name}
    # Give us all the programs installed if user does not specify any programs
    if (-not $Programs) {
        $OutputList = $ProgramsList
    }
    
    return $OutputList
}

function Assert-RPOSInstall {
    [CmdletBinding()]
    param (
        [Parameter()]
        [Version]
        $RPOSTargetVersion
    )

    # Define registry key to check
    $RegKey = "HKLM:\Software\Radiant\RPOS\Somewhere\Version"  
    if ($Config.RPOSVersionRegistryKey) {
        $RegKey = $Config.RPOSVersionRegistryKey
    }

    $Version = [version](Get-ItemPropertyValue -Path $RegKey)                                                                                           # TODOOOOOOOOO

    if ($Version -eq $RPOSTargetVersion) {
        return $true
    }

    return $false
}

function Assert-IsAdministrator {
    # Blatently copied from https://github.com/cloudbase/windows-imaging-tools/blob/master/WinImageBuilder.psm1#L106-L114
    $UserID = [System.Security.Principal.WindowsIdentity]::GetCurrent()
    $WindowsPrincipal = New-Object System.Security.Principal.WindowsPrincipal($UserID)
    $AdminGroupMembership = [System.Security.Principal.WindowsBuiltInRole]::Administrator

    $IsAdministrator = $WindowsPrincipal.IsInRole($AdminGroupMembership)
    if (!($IsAdministrator)) {
        throw [System.AccessViolationException]
    }
}

function Install-RPOS {
    [CmdletBinding()]
    param (
        [Parameter()]
        [Version]
        $RPOSVersion=$Config.TargetRPOSVersion
    )
    # Check if the function was run with Admin privs
    try {
        Assert-IsAdministrator # Will throw the caught expection type System.AccessViolationException
        Write-Log -Level "Info" -Message "Starting RPOS v.$RPOSVersion Installer" -Thread $MyInvocation.MyCommand.Name
        Start-ExeInstaller -RPOSSetupPath $Config.DefaultSetupDir
    }
    catch [System.AccessViolationException] {
        Write-Log -Message "RPOS v.$RPOSVersion Installer should be run with an Administrator account." -Thread $MyInvocation.MyCommand.Name
    }
    finally {
        Export-Log
    } 
    
}


Export-ModuleMember -Function Confirm-PathExists, Write-Log, Export-Log, Start-ExeInstaller, Get-InstalledSoftware, Assert-RPOSInstall, Assert-IsAdministrator, Install-RPOS